"""Utility classes for the building block 'selector'."""
from typing import List, Callable, Dict, Any, Iterable

from .models import UDFModel
from .typing import SelectorExpression, Payload
from .utils import Singleton, safe_eval, FallbackBox, EvaluationError
from .validator import Validator


class PayloadSelector(Singleton):
    """
    The actual selector implementation.
    Knows about globals, the suppress literal, udfs and how to evaluate
    simple and complex selector expressions.
    """

    def __init__(self: 'PayloadSelector') -> None:  # pylint: disable=super-init-not-called
        self._suppress_literal = object()
        self._custom = {}  # type: Dict[str, Callable[..., Any]]
        self._register_globals()

    @property
    def suppress(self) -> object:
        """Return the suppress literal."""
        return self._suppress_literal

    @property
    def suppress_aliases(self) -> List[str]:
        """Returns available aliases for the suppress literal."""
        uppers = ["SUPPRESS", "SUPPRESSME", "SUPPRESSPUSH", "SUPPRESS_ME", "SUPPRESS_PUSH"]
        return uppers + [item.lower() for item in uppers]

    def _register_globals(self) -> None:
        self._custom["abs"] = abs
        self._custom["bool"] = bool
        self._custom["dict"] = dict
        self._custom["float"] = float
        self._custom["getattr"] = getattr
        self._custom["hasattr"] = hasattr
        self._custom["hash"] = hash
        self._custom["int"] = int
        self._custom["isinstance"] = isinstance
        self._custom["len"] = len
        self._custom["list"] = list
        self._custom["max"] = max
        self._custom["min"] = min
        self._custom["ord"] = ord
        self._custom["pow"] = pow
        self._custom["reversed"] = reversed
        self._custom["round"] = round
        self._custom["sorted"] = sorted
        self._custom["str"] = str
        self._custom["zip"] = zip

        from base64 import b64encode, b64decode
        self._custom["b64encode"] = b64encode
        self._custom["b64decode"] = b64decode
        import os
        self._custom["basename"] = os.path.basename
        from .utils import on_off
        self._custom["on_off"] = on_off

    def register_custom_global(self, name: str, fun: Callable[..., Any]) -> None:
        """
        Register a custom function that should be available in the context of the selector.

        Args:
            name: Name of the function to register.
            fun: The actual function to register.

        Returns:
            None.
        """
        Validator.is_function(fun=fun)
        Validator.is_instance(str, name=name)
        if name not in self._custom:  # Overriding a already existing custom will silently fail!
            self._custom[name] = fun

    def register_udfs(self, udfs: Iterable[UDFModel]) -> None:
        """Register the given user-definied function."""
        Validator.is_iterable_but_no_str(udfs=udfs)
        for i, udf in enumerate(udfs):
            Validator.is_instance(UDFModel, **{'udfs.item_{i}'.format(i=i): udf})  # type: ignore
            self.register_custom_global(udf.name, udf.callable)

    def _eval_wrapper(self, selector: str, payload: Payload) -> Payload:
        suppress_kwargs = {alias: self.suppress for alias in self.suppress_aliases}
        return safe_eval(
            source=selector,
            payload=payload,
            data=payload,
            **suppress_kwargs,
            **self._custom
        )

    @staticmethod
    def _isalambda(v: Any) -> bool:
        lambda_ = lambda: 0
        return isinstance(v, type(lambda_)) and v.__name__ == lambda_.__name__

    def _e(self, snippet: SelectorExpression, payload: Payload) -> Payload:
        # There might be nested dicts or lists inside the complex structure... Eval them recursively
        if isinstance(snippet, dict):
            return {self._e(k, payload): self._e(v, payload) for k, v in snippet.items()}
        if isinstance(snippet, list):
            return [self._e(i, payload) for i in snippet]

        # Test if the snippet constructs a lambda
        # -> assumes to be callable code with payload argument
        try:
            possible_fun = self._eval_wrapper(str(snippet), payload)
        except EvaluationError:
            if str(snippet).startswith('lambda'):
                raise EvaluationError("Your lambda is errorneous: '{snippet}'".format(
                    **locals()))
            # Evaluation failed -> assume that is string literal (dict key/value or list item)
            return snippet

        if not self._isalambda(possible_fun):
            return snippet  # Cannot be executed. So assume it is not a selector expression

        # It is a callable. Lets modify globals and call it :-)
        suppress_kwargs = {alias: self.suppress for alias in self.suppress_aliases}
        customs = self._custom
        for k, v in {**suppress_kwargs, **customs}.items():
            possible_fun.__globals__[k] = v
        try:
            return possible_fun(payload)
        except:
            raise EvaluationError("Error when running the selector lambda: '{snippet}'".format(
                **locals()))

    def eval_selector(self, selector: SelectorExpression, payload: Payload) -> Payload:
        """Applies the specified selector to the given payload."""
        # Wrap payload inside a Box -> this makes dot accessable dictionaries possible
        # We create a dictionary cause payload might not be an actual dictionary.
        # This way wrapping with Box will always work ;-)
        payload = FallbackBox({'base': payload}).base
        if selector is None:
            return payload

        if isinstance(selector, (list, dict)):
            return self._e(selector, payload)  # Complex. Need additional magic

        # No complex structure. We assume that is an expression and we try to evaluate it
        return self._eval_wrapper(str(selector), payload)



"""Basic stuff for implementing pull plugins."""

import multiprocessing as proc
import time
from abc import abstractmethod
from datetime import datetime
from typing import Any, Callable, Optional

from schedule import Scheduler  # type: ignore

from .. import Plugin
from ...typing import Payload
from ...utils import (auto_str_ignore, parse_duration_literal, StopCycleError,
                      interruptible_sleep, try_parse_bool, DurationLiteral)


@auto_str_ignore(['stopped', '_stopped', 'on_payload'])
class PullBase(Plugin):
    """
    Base class for pull plugins.
    The plugin has to implements `pull` to do the actual data retrieval / task.
    """
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._stopped = proc.Event()

    @property
    def stopped(self) -> bool:
        """Returns True if the pull is considered stopped; otherwise False."""
        return self._stopped.is_set()

    @abstractmethod
    def pull(self) -> None:
        """
        Performs the actual data retrieval / task.
        Returns:
            None.
        """
        raise NotImplementedError()  # pragma: no cover

    def on_payload(self, payload: Payload) -> None:
        """Callback for execution engine."""

    def notify(self, payload: Payload) -> None:
        """Call in subclass to emit some payload to the execution engine."""
        self.on_payload(self, payload)  # type: ignore  # pylint: disable=too-many-function-args

    def stop(self) -> None:
        """Stops the plugin."""
        self._stopped.set()

    def _sleep(self, sleep_time: float = 10) -> None:
        """Call in subclass to perform some sleeping."""
        def callback() -> None:
            if self.stopped:
                raise StopCycleError()
        interruptible_sleep(sleep_time, callback, interval=0.5)


class PollingError(Exception):
    """Base class for errors that occur during polling."""


class StopPollingError(Exception):
    """Raise this exception in child classes if you want to stop the polling scheduler."""


@auto_str_ignore(['_scheduler'])
class Polling(PullBase):
    """
    Base class for polling plugins.

    You may specify duration literals such as 60 (60 secs), 1m, 1h (...) to realize a periodic
    polling or cron expressions (*/1 * * * * > every min) to realize cron like behaviour.
    """
    __prefix__ = 'poll'

    def __init__(self, interval: DurationLiteral = 60, instant_run: bool = False, **kwargs: Any):
        super().__init__(**kwargs)

        try:
            # Literals such as 60s, 1m, 1h, ...
            self._poll_interval = parse_duration_literal(interval)
            self.is_cron = False
        except TypeError:
            # ... or a cron-like expression is valid
            from cronex import CronExpression  # type: ignore
            self._cron_interval = CronExpression(interval)
            self.is_cron = True

        self._scheduler = None  # type: Optional[Scheduler]
        self._instant_run = try_parse_bool(instant_run, False)

    def pull(self) -> None:
        self._scheduler = Scheduler()
        self._configure_scheduler(self._scheduler, self._run_schedule)

        if self._instant_run:
            self._scheduler.run_all()

        while not self.stopped:
            self._scheduler.run_pending()
            time.sleep(0.5)

    def _run_schedule(self) -> None:
        try:
            if self.is_cron:
                dtime = datetime.now()
                if not self._cron_interval.check_trigger((dtime.year, dtime.month,
                                                          dtime.day, dtime.hour, dtime.minute)):
                    return  # It is not the time for the cron to trigger
            payload = self.poll()
            if payload is not None:
                self.notify(payload)
        except StopPollingError:
            self.stop()
        except Exception:  # pragma: no cover, pylint: disable=broad-except
            import traceback
            self.logger.error("\n%s", traceback.format_exc())

    def _configure_scheduler(self, scheduler: Scheduler, callback: Callable[[], None]) -> None:
        """
        Configures the scheduler. You have to differ between "normal" intervals and
        cron like expressions by checking `self.is_cron`.

        Override in subclasses to fir the behaviour to your needs.
        Args:
            scheduler (schedule.Scheduler): The actual scheduler.
            callback (callable): The callback to call when the time is right.

        Returns:
            None

        """
        if self.is_cron:
            # Scheduler always executes at the exact minute to check for cron triggering
            scheduler.every().minute.at(":00").do(callback)
        else:
            # Scheduler executes every interval seconds to execute the poll
            scheduler.every(self._poll_interval).seconds.do(callback)

    @abstractmethod
    def poll(self) -> None:
        """
        Implement in plugin components to do the actual polling.

        Returns: Returns the data for the downstream pipeline.

        """
        raise NotImplementedError()  # pragma: no cover



"""Z-Way related pulls."""

import re

import requests
from requests.auth import HTTPBasicAuth
import schema

from . import Polling, PollingError
from .http import Server
from ...utils import auto_str_ignore, try_parse_int_float_str
from ...validator import Validator


@auto_str_ignore(['password'])
class ZwayPoll(Polling):
    """
    Pulls the specified json content from the z-way rest api.
    The content is specified by the url, e.g. http://<host>:8083/ZWaveAPI/Run/devices will pull
    all devices and serve the result as a json.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/zway.ZwayPoll/index.md
    """

    __prefix__ = 'zway'

    def __init__(self, url, user, password, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.user = str(user)
        self.password = str(password)

    def poll(self):
        self.logger.debug("Polling url '%s'", self.url)
        reply = requests.get(self.url, auth=HTTPBasicAuth(self.user, self.password))
        if reply.status_code != requests.codes.ok:  # pylint: disable=no-member
            raise PollingError(
                "Code {code}: {error}".format(code=reply.status_code, error=reply.text)
            )

        try:
            return reply.json()
        except ValueError:
            # No valid json, try text
            return reply.text


class ZwayReceiver(Server):
    """
    Setups a http server to process incoming GET-requests from the
    Z-Way-App [`HttpGet`](https://github.com/hplato/Zway-HTTPGet/blob/master/index.js).

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/pull/zway.ZwayReceiver/index.md
    """
    MAPPING_SCHEMA = schema.Schema({
        schema.Optional(str): schema.Or(
            {
                'alias': schema.Use(str),
                schema.Optional(str): object
            },
            schema.Use(lambda val: dict(alias=val, props={}))
        )
    })

    DEVICE_REGEX = r'[a-zA-Z0-9_\-]+'
    VALUE_REGEX = r'[a-zA-Z0-9\.,]+'
    MODE_AUTO = 'auto'
    MODE_MAPPING = 'mapping'
    MODE_BOTH = 'both'
    VALID_MODES = [MODE_AUTO, MODE_MAPPING, MODE_BOTH]
    VDEV_REGEX = r'ZWayVDev_zway_(?P<device_id>[0-9]+)\-(?P<instance>[0-9]+)\-(?P<cclass>[0-9]+)' \
                 r'(\-(?P<mode>[0-9]+))?'

    AUTO_MAPPING = {
        '37': 'switch',  # Switch binary
        '38': 'level',  # Switch multilevel
        '48': {'1': 'motion'},  # Sensor binary
        '49': {'1': 'temperature', '3': 'illumination', '4': 'power'},  # Sensor multilevel
        '50': {'0': 'consumption'},  # Meter
        '67': {'1': 'setpoint'},  # Thermostat
        '128': 'battery'
    }

    @staticmethod
    def _make_regex(url_format):
        escaped_url = re.escape(url_format)
        device_group = "(?P<device>{})".format(ZwayReceiver.DEVICE_REGEX)
        value_group = "(?P<value>{})".format(ZwayReceiver.VALUE_REGEX)
        return re.compile(
            '.*' + escaped_url.replace(r'\%DEVICE\%', device_group)
            .replace(r'\%VALUE\%', value_group)
            .replace(r'%DEVICE%', device_group)
            .replace(r'%VALUE%', value_group)
        )

    @staticmethod
    def _safe_get_group(match, group_name):
        try:
            return match.group[group_name]
        except IndexError:
            return None

    @classmethod
    def _auto_map(cls, cclass, mode):
        descr = cls.AUTO_MAPPING.get(str(cclass))
        if not descr:
            return None
        if isinstance(descr, dict) and mode:
            descr = descr.get(str(mode))
        return descr

    def __init__(self, url_format="/set?device=%DEVICE%&state=%VALUE%", device_mapping=None,
                 ignore_unknown_devices=False, mode='mapping', vdev_regex=None, **kwargs):
        # Zway module HttpGet only supports GET calls
        kwargs['allowed_methods'] = 'GET'
        super().__init__(**kwargs)
        self.ignore_unknown_devices = bool(ignore_unknown_devices)

        Validator.is_instance(str, url_format=url_format)
        if not url_format.startswith('/'):
            url_format = '/' + url_format
        if '?' not in url_format:
            url_format = url_format + '?'
        self.url_format = url_format
        self.url_format_regex = self._make_regex(url_format)

        if device_mapping is None:
            device_mapping = {}
        self.device_mapping = self.MAPPING_SCHEMA.validate(device_mapping)

        Validator.one_of(self.VALID_MODES, False, mode=mode)
        self.mode = mode

        if mode in [self.MODE_BOTH, self.MODE_AUTO]:
            self._vdev_regex = vdev_regex or self.VDEV_REGEX
            self._vdev_regex_compiled = re.compile(self.VDEV_REGEX)

    def notify(self, payload):
        full_path = payload['full_path']
        match = self.url_format_regex.match(full_path)
        if match is None:
            self.logger.warning(
                "Cannot extract device_name and/or value from '%s' with regex '%s'",
                full_path, self.url_format_regex
            )
            return
        raw_device = match.group('device')
        value = match.group('value')

        device_map = None
        if self.mode in [self.MODE_MAPPING, self.MODE_BOTH]:
            device_map = self.device_mapping.get(raw_device)

        if not device_map and self.mode in [self.MODE_AUTO, self.MODE_BOTH]:
            match = self._vdev_regex_compiled.match(raw_device)

            if match is not None:
                device_id, cclass, mode = (
                    match.group('device_id'),
                    match.group('cclass'),
                    match.group('mode')
                )
                type_of_device = self._auto_map(cclass, mode)
                if not type_of_device and self.ignore_unknown_devices:
                    self.logger.debug(
                        "Got device '%s' but it is unknown to the automapper", raw_device
                    )
                    return
                device_map = {
                    'alias': device_id,
                    'command_class': cclass,
                    'mode': mode,
                    'type': type_of_device or (cclass + '-{}'.format(mode) if mode else '')
                }
            else:
                self.logger.warning(
                    "Cannot extract device_name and/or value from '%s' with regex '%s'",
                    raw_device, self._vdev_regex
                )

        if device_map is None:
            if self.ignore_unknown_devices:
                self.logger.debug("Got device '%s' but it is ignored", raw_device)
                return  # Device is not mapped and should be ignored-> abort
            device_map = dict(alias=raw_device)  # If it isn't ignored just take the raw name

        device_map = device_map.copy()
        alias = device_map.pop('alias')
        super().notify(dict(device_name=alias, props=device_map, raw_device=raw_device,
                            value=try_parse_int_float_str(value)))



"""MQTT related utility classes."""

import json
import logging

from ..utils import auto_str_ignore, try_parse_int

_LOGGER = logging.getLogger(__name__)


@auto_str_ignore(['password'])
class MQTTBase:
    """MQTT base class for publising message to a mqtt broker."""

    def __init__(self, host, port=1883, user=None, password=None, qos=0, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.port = int(port)
        self.user = user and str(user)
        self.password = password and str(password)
        self.qos = self._parse_qos(qos)

    @staticmethod
    def _parse_qos(val):
        pval = try_parse_int(val)
        if pval is None:
            _LOGGER.warning("QOS level of '%s' is not int parsable. Defaulting to 0", val)
            return 0
        if pval < 0:
            _LOGGER.warning("QOS level of '%s' is invalid. Defaulting to 0", val)
            return 0
        if pval > 2:
            _LOGGER.warning("QOS level of '%s' is invalid. Defaulting to 2", val)
            return 2
        return pval

    def _publish(self, payload, topic, retain=False, qos=None):
        if isinstance(payload, (dict, list, tuple)):
            payload = json.dumps(payload)

        auth = None
        if self.user:
            auth = dict(username=self.user, password=self.password)

        if qos is None:
            qos = self.qos

        import paho.mqtt.publish as publish
        publish.single(
            topic=topic,
            payload=payload,
            hostname=self.host,
            port=self.port,
            retain=retain,
            auth=auth,
            qos=qos
        )
        _LOGGER.debug(
            "Published message on '%s' @ %s:%s with qos=%s. Payload='%s'",
            topic, self.host, self.port, qos, payload
        )



"""Some typing stuff."""

from typing import Any, Dict, Union, List, Callable, Optional

from typing_extensions import Protocol

AnyCallable = Callable[..., Any]
DurationLiteral = Union[str, int]
Envelope = Dict[str, Any]
Payload = Any  # pylint: disable=invalid-name
SelectorExpression = Union[str, Dict[str, str], List[str]]


class QueuePutGet(Protocol):
    """Queue that supports put and get."""
    def get(self) -> Any:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...

    def put(self, item: Any) -> None:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...


class StopSignal(Protocol):
    """Stopping signal."""
    def set(self) -> None:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...

    def is_set(self) -> bool:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...


class ThreadLike(Protocol):
    """Thread like interface."""
    def is_alive(self) -> bool:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...

    def join(self, timeout: Optional[float]) -> None:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...

    def start(self) -> None:  # pylint: disable=unused-argument,missing-docstring,no-self-use
        ...



"""Engine implementation using multiprocessing."""

import multiprocessing as proc
import os
from typing import cast

from .parallel import (StoppableRunner, StoppableWorker, ParallelEngine, RetryHandler,
                       QueuePutGet, StopSignal)
from ..models import TaskModel


class ProcessStoppableRunner(StoppableRunner):
    """Runner implementation using multiprocessing thread-like interface."""
    @staticmethod
    def _make_shutdown_event() -> StopSignal:
        return proc.Event()

    @staticmethod
    def get_ident() -> str:
        return str(os.getppid())

    def start(self) -> None:
        if self._runner:
            raise RuntimeError("Runner is already started and cannot restart...")
        self._runner = proc.Process(target=self._start_pull)
        self._runner.start()

    def terminate(self) -> None:
        self._assert_runner()
        self._runner.terminate()  # type: ignore


class ProcessWorker(StoppableWorker):
    """Worker implementation using multiprocessing thread-like interface."""
    @staticmethod
    def get_ident() -> str:
        return str(os.getppid())

    def start(self) -> None:
        if self._worker:
            raise RuntimeError("Worker is already started and cannot restart...")
        self._worker = proc.Process(target=self._process_queue)
        self._worker.start()

    def terminate(self) -> None:
        self._worker.terminate()  # type: ignore


class ProcessEngine(ParallelEngine):
    """Engine implementation using multiprocessing thread-like interface."""
    def _make_worker(self) -> ProcessWorker:
        return ProcessWorker(self.queue, self.stop_working_item)

    def _make_runner(self, task: TaskModel, retry_handler: RetryHandler) -> ProcessStoppableRunner:
        return ProcessStoppableRunner(task, self.queue, retry_handler=retry_handler)

    def _make_queue(self) -> QueuePutGet:
        return cast(QueuePutGet, proc.Queue())

    def _make_shutdown_event(self) -> StopSignal:
        return proc.Event()

    def _stop_worker(self) -> None:
        # We do not use the stop_working_item here... We ungentle terminate the worker
        for worker in self.worker:
            self.logger.debug("Terminating worker: %s", worker)
            worker.terminate()



"""Engine implementation using threads."""

from .parallel import ParallelEngine


# Re-use the ParallelEngine which is already implemented via threads.
class ThreadEngine(ParallelEngine):
    """Engine implementation using threads."""



"""Contains base stuff for user-defined functions in selector expressions."""

from abc import abstractmethod
from datetime import datetime, timedelta
from typing import Optional, Any

from .. import Plugin
from ...utils import parse_duration_literal, DurationLiteral


class UserDefinedFunction(Plugin):
    """Base class for a user defined expression."""

    def __init__(self, throttle: Optional[DurationLiteral] = None, **kwargs: Any):
        """
        Initializer.

        Args:
            throttle: If set to a valid duration literal (e.g. 5m) the return value of the
                called functions will be cached for the amount of time.
        """
        super().__init__(**kwargs)
        self.throttle = throttle and parse_duration_literal(throttle)
        self._cache = None
        self._last_call = None  # type: Optional[datetime]

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if not self.throttle:
            return self.action(*args, **kwargs)

        # Invariant: self.throttle is not None
        now = datetime.now()
        if self._last_call is None:
            res = self.action(*args, **kwargs)
            self._cache = res
            self._last_call = now
            return res

        # Invariant: self._last_call is not None
        span = now - self._last_call
        if span >= timedelta(seconds=self.throttle):
            res = self.action(*args, **kwargs)
            self._cache = res
            self._last_call = now
            return res

        # Invariant: span < delta(throttle)
        return self._cache

    @abstractmethod
    def action(self, *args: Any, **kwargs: Any) -> Any:
        """Actual definition of the hard-work of the user defined function."""
        raise NotImplementedError()



"""Base implementation for parallel engines."""

import copy
import multiprocessing
import threading
import time
from queue import Queue
from typing import Any, Optional, cast, List

from . import RetryHandler, PushExecutor, Engine, SimpleRetryHandler
from ..models import TaskModel, TaskSet
from ..typing import Payload, QueuePutGet, StopSignal, ThreadLike
from ..utils import Loggable, auto_str, auto_str_ignore, sleep_until_interrupt
from ..validator import Validator


@auto_str(__repr__=True)
@auto_str_ignore(ignore_list=["queue", "stopped", "_runner"])
class StoppableRunner(Loggable):
    """
    Base implementation of a parallel runner that executes a `pull`. By default threading is used.

    For other implementations you have basically to override / adapt:
    * _make_shutdown_event() -> Create an event that can signal if this task should stop. The event
        should provide `is_set()` and `set()` methods (like the `Event` from `threading` does).
    * get_ident() -> Some unique identifier
    * start() -> Starts the runner. Start is async and will not block the current execution.
    * join(timeout=...) -> Wait for the runner to finish. Will block!
    * is_alive() -> Checks if the runner is currently running
    * terminate() -> Terminate the runner. Will only be called when the runner does not stop
        after `stop()` was called.
    """

    def __init__(self, task: TaskModel, queue: QueuePutGet, retry_handler: RetryHandler):
        """
        Initializer.
        Args:
            task: Task to execute.
            queue: Queue to use for communication between runner and worker.
                Should provide `put(item)` method.
            retry_handler: Retry handler.
        """
        Validator.is_instance(TaskModel, task=task)
        self.task = task
        self.queue = queue
        self.stopped = self._make_shutdown_event()
        Validator.is_instance(RetryHandler, retry_handler=retry_handler)
        self.retry_handler = retry_handler
        self._runner = None  # type: Optional[ThreadLike]

    @staticmethod
    def _make_shutdown_event() -> StopSignal:
        """Create an event that can signal if this task should stop. The event should provide
        `is_set()` and `set()` methods (like the `Event` from `threading` does)."""
        return threading.Event()

    @staticmethod
    def get_ident() -> str:
        """Return some unique identifier for this runner."""
        return str(threading.get_ident())

    def start(self) -> None:
        """Start the runner and the processing of the enveloped task."""
        if self._runner:
            raise RuntimeError("Runner is already started and cannot restart...")
        self._runner = threading.Thread(target=self._start_pull)
        self._runner.start()

    def join(self, timeout: Optional[float] = None) -> None:
        """Wait for the runner to exit."""
        self._assert_runner()
        cast(threading.Thread, self._runner).join(timeout=timeout)

    def is_alive(self) -> bool:
        """Check if the runner is still alive."""
        self._assert_runner()
        return cast(threading.Thread, self._runner).is_alive()

    def terminate(self) -> None:
        """Terminates the runner."""
        self.logger.warning(
            "Method terminate() on basic StoppableRunner is not supported.\n"
            "The thread will NOT terminate."
        )

    def _assert_runner(self) -> None:
        if not self._runner:
            raise RuntimeError("Runner is is not started")

    def _start_pull(self) -> None:
        def on_payload(plugin: Any, payload: Payload) -> None:  # pylint: disable=unused-argument
            for push in self.task.pushes:
                self.logger.debug(
                    "[Task-%s] Queing item '%s' for push '%s'",
                    self.get_ident(),
                    payload,
                    push
                )
                self.queue.put((payload, push))

        self.task.pull.instance.on_payload = on_payload  # type: ignore

        while not self.stopped.is_set():
            try:
                self.task.pull.instance.pull()
                if not self.stopped.is_set():
                    # Bad thing... Pulling exited unexpectedly
                    self.logger.error(
                        "[Task-%s] Pulling of '%s' exited unexpectedly",
                        self.get_ident(),
                        self.task.pull.instance,
                    )
            except KeyboardInterrupt:  # pragma: no cover
                self.logger.debug(
                    "[Task-%s] Pulling of '%s' hit a keyboard interrupt",
                    self.get_ident(),
                    self.task.pull.instance
                )
            except:  # pylint: disable=bare-except
                # Bad thing... Pulling exited with exception
                import traceback
                self.logger.error(
                    "[Task-%s] Pulling of '%s' raised an error\n%s",
                    self.get_ident(),
                    self.task.pull.instance,
                    traceback.format_exc()
                )
            finally:
                self._handle_pull_exit()

    def _handle_pull_exit(self) -> None:
        if not self.stopped.is_set():
            directive = self.retry_handler.handle_error()
            if directive.abort:
                self.logger.error(
                    "[Task-%s] Pulling of '%s' exited due to retry limitation",
                    self.get_ident(),
                    self.task.pull.instance,
                )
                self.stopped.set()
            else:
                sleep_until_interrupt(directive.wait_for, self.stopped.is_set)

    def stop(self) -> None:
        """Stop the runner gracefully."""
        self.logger.info(
            "[Task-%s] Got stopping signal: '%s'",
            self.get_ident(),
            self.task.name
        )
        self._assert_runner()
        self.stopped.set()
        self.task.pull.instance.stop()


@auto_str(__repr__=True)
class StoppableWorker(Loggable):
    """
    Base implementation of a worker that processes fetches items from the queue and delegates it to
    a push for further execution. By default threading is used.

    To stop the worker you should put the `stop_working_item` on the queue.

    For other implementations you have basically to override / adapt:
    * get_ident(): Some unique identifier
    * start(): Starts the worker. Start is async and will not block the current execution.
    * join(timeout=...): Wait for the worker to finish. Will block!
    * is_alive(): Checks if the worker is currently running
    * terminate(): Terminate the worker. Will only be called when the worker does not stop after
        receiving the `stop_working_item`.
    """
    def __init__(self, queue: QueuePutGet, stop_working_item: Any):
        """
        Initializer.

        Args:
            queue: Queue to fetch processable items from. Should provide `get()` method.
            stop_working_item: Stop the worker when this item hits the queue.
        """
        self.queue = queue
        self.stop_working_item = stop_working_item
        self._worker = None  # type: Optional[ThreadLike]

    @staticmethod
    def get_ident() -> str:
        """Return a unique identifier for this worker."""
        return str(threading.get_ident())

    def start(self) -> None:
        """Start this worker and begin to pull from the queue."""
        if self._worker:
            raise RuntimeError("Worker is already started and cannot restart...")
        self._worker = threading.Thread(target=self._process_queue)
        self._worker.start()

    def join(self, timeout: Optional[float] = None) -> None:
        """Wait for the worker to exit."""
        self._assert_runner()
        cast(threading.Thread, self._worker).join(timeout=timeout)

    def is_alive(self) -> bool:
        """Check if the worker is still running and alive."""
        self._assert_runner()
        return cast(threading.Thread, self._worker).is_alive()

    def terminate(self) -> None:
        """Termintes the worker forcefully."""
        self.logger.warning("Method terminate() on basic StoppableWorker is not supported.\n"
                            "The thread will NOT terminate.")

    def _assert_runner(self) -> None:
        if not self._worker:
            raise RuntimeError("Worker is is not started")

    def _process_queue(self) -> None:
        while True:
            try:
                payload = self.queue.get()
                if payload is self.stop_working_item:
                    self.logger.info("[Worker-%s] Got stopping signal", self.get_ident())
                    # Notify other worker as well
                    self.queue.put(self.stop_working_item)
                    break
                try:
                    # We assume payload is actual payload, push
                    payload, push = payload
                    PushExecutor().execute(
                        ident=str(self.get_ident()),
                        payload=payload,
                        push=push,
                        result_callback=lambda res, dep: self.queue.put((res, dep))
                    )
                finally:
                    # self.queue.task_done()
                    pass
            except KeyboardInterrupt:  # pragma: no cover
                pass
            except Exception:  # pragma: no cover, pylint: disable=broad-except
                import traceback
                self.logger.error("\n%s", traceback.format_exc())


@auto_str_ignore(ignore_list=["stop_working_item", "queue", "shutdown", "runner", "worker"])
class ParallelEngine(Engine):
    """Base implementation of a parallel engine.
    Spawns runner to process tasks and workers to handle any emitted results from `pull`
    components."""

    def __init__(self, queue_worker: int = 3, retry_handler: Optional[RetryHandler] = None,
                 stop_timeout: Optional[int] = 5):
        super().__init__()
        self.queue_worker = max(int(queue_worker), 1)
        self.retry_handler = retry_handler
        if self.retry_handler is None:
            self.retry_handler = SimpleRetryHandler()
        Validator.is_instance(RetryHandler, retry_handler=self.retry_handler)
        self.stop_working_item = object()
        self.queue = self._make_queue()
        self.runner = []  # type: List[StoppableRunner]
        self.worker = []  # type: List[StoppableWorker]
        self.shutdown = self._make_shutdown_event()
        self.stop_timeout = Validator.cast_or_none(int, stop_timeout) or 5

    def _make_worker(self) -> StoppableWorker:
        return StoppableWorker(self.queue, self.stop_working_item)

    def _make_runner(self, task: TaskModel, retry_handler: RetryHandler) -> StoppableRunner:
        return StoppableRunner(task, self.queue, retry_handler=retry_handler)

    def _make_queue(self) -> QueuePutGet:  # pylint: disable=no-self-use
        return cast(QueuePutGet, Queue())

    def _make_shutdown_event(self) -> StopSignal:  # pylint: disable=no-self-use
        # We might run the Engine in an isolated process.
        # So the multiprocessing.Event() is better suited.
        return multiprocessing.Event()

    def _run_queue_worker(self) -> None:
        n_worker = self.queue_worker
        for i in range(n_worker):
            thr = self._make_worker()
            thr.start()
            self.logger.info("[Worker-%s] Started (%s/%s)", thr.get_ident(), i + 1, n_worker)
            self.worker.append(thr)

    def _run_tasks(self, tasks: TaskSet) -> None:
        for _, task in tasks.items():
            thr = self._make_runner(task, copy.deepcopy(cast(RetryHandler, self.retry_handler)))
            thr.start()
            self.logger.info("[Task-%s] Started for task '%s'", thr.get_ident(), task.name)
            self.runner.append(thr)

    def _stop_runner(self) -> None:
        for runner in self.runner:
            self.logger.debug("Stopping runner: %s", runner)
            runner.stop()

    def _stop_worker(self) -> None:
        self.logger.debug("Pushing stop_working_item on queue")
        self.queue.put(self.stop_working_item)

    def _stop_all(self) -> None:
        self._stop_runner()
        self._stop_worker()

    def _wait_for_all(self) -> None:
        for thr in self.runner + self.worker:  # type: ignore
            thr.join(timeout=self.stop_timeout)
            if thr.is_alive():
                self.logger.warning("%s did not respond to stop command. Terminating...", thr)
                thr.terminate()

    def _run(self, tasks: TaskSet) -> None:
        for _, thr in tasks.items():
            if not isinstance(thr, TaskModel):
                raise TypeError("All items of argument 'tasks' are expected to be an instance "
                                "of 'Task'")

        self._run_queue_worker()
        self._run_tasks(tasks)
        try:
            while not self.shutdown.is_set():
                time.sleep(1)
        except KeyboardInterrupt:  # pragma: no cover
            pass
        self.logger.debug("Exited control loop. Stopping all threads gracefully")
        self._stop_all()
        self._wait_for_all()

    def _stop(self) -> None:
        self.shutdown.set()



"""Contains base classes for engines."""

import copy
from abc import abstractmethod
from datetime import datetime
from typing import Any, Callable, Optional

import attr

from ..models import TaskSet, PushModel
from ..selector import PayloadSelector
from ..typing import Payload
from ..utils import (Loggable, Singleton, parse_duration_literal, DurationLiteral, auto_str,
                     is_iterable_but_no_str)
from ..validator import Validator


class NotSupportedError(Exception):
    """Is raised when a task is not supported by an engine."""


PushResultCallback = Callable[[Payload, PushModel], None]


@auto_str(__repr__=True)
class Engine(Loggable):
    """
    An engine is responsible for launching a defined set of tasks and the communication between
    pulls and and their associated pushes.

    A call to to an engine's `run(...)` method will block the calling thread until the engine
    decides the job is done (normally an external SIGTERM occurs)
    """
    def __init__(self: 'Engine'):
        pass

    def run(self, tasks: TaskSet) -> None:
        """Run the given task set inside the engine."""
        return self._run(tasks)

    @abstractmethod
    def _run(self, tasks: TaskSet) -> None:
        """Run the given task set inside the engine. Override in child classes to do the hard
        work."""
        raise NotImplementedError()

    def stop(self) -> None:
        """Stop the engine."""
        self._stop()

    @abstractmethod
    def _stop(self) -> None:
        """Stop the engine. Override in child classes."""
        raise NotImplementedError()


@attr.s
class RetryDirective:
    """Contains directive information for engines on how to proceed in erroneous cases."""

    # If set to True, instructs the engine to abort the pull; otherwise retry the pull
    abort = attr.ib(converter=bool, type=bool, default=True)  # type: bool
    # Instructs the engine to wait some time before retrying the pull again
    wait_for = attr.ib(converter=parse_duration_literal, type=int, default=0)  # type: int
    # Just contextual information how many retries occurred so far
    retry_cnt = attr.ib(converter=int, type=int, default=0)  # type: int


@auto_str(__repr__=True)
class RetryHandler:
    """
    A retry handler comes into play when a pull component exits unexpectedly whether by breaking
    the actual loop or by error. The `RetryHandler` decides how to proceed further (wait and retry
    the pull or more sophisticated logic).
    """
    def __init__(self, **kwargs: Any):
        pass

    @abstractmethod
    def handle_error(self) -> RetryDirective:
        """
        Asks the retry handler to handle an error.

        Returns:
            An instance of `RetryDirective`.
        """
        raise NotImplementedError()


class NoRetryHandler(RetryHandler):
    """Will instruct the engine to abort and not to retry the pull."""

    def handle_error(self) -> RetryDirective:
        return RetryDirective(abort=True, wait_for=0, retry_cnt=1)


class SimpleRetryHandler(RetryHandler):
    """Simply instructs the engine to wait for the given amount of time after an error."""

    def __init__(self, retry_wait: DurationLiteral = 60, **kwargs: Any):
        super().__init__(**kwargs)
        self.retry_wait = parse_duration_literal(retry_wait)
        self.retry_count = 0

    def _incr_retry(self) -> None:
        self.retry_count += 1

    def handle_error(self) -> RetryDirective:
        self._incr_retry()
        return RetryDirective(abort=False, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class LimitedRetryHandler(SimpleRetryHandler):
    """Instructs the engine to wait for the given amout of time after an error. If the specified
     `max_retries` is hit the engine is instructed to abort the `pull`."""

    def __init__(self, max_retries: Optional[int] = 3, **kwargs: Any):
        super().__init__(**kwargs)
        self.max_retries = Validator.cast_or_none(int, max_retries)  # type: Optional[int]

    def _eval_abort(self, retry_count: int) -> bool:
        if self.max_retries is None or self.max_retries < 0:
            return False
        return retry_count > self.max_retries

    def handle_error(self) -> RetryDirective:
        super().handle_error()
        abort = self._eval_abort(self.retry_count)
        return RetryDirective(abort=abort, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class AdvancedRetryHandler(LimitedRetryHandler):
    """Works like the `LimitedRetryHandler` but will reset the retry count when a given amount of
    time between the current failure and the previous failure has passed."""

    def __init__(self, reset_retry_threshold: DurationLiteral = 60, **kwargs: Any):
        super().__init__(**kwargs)
        # Reset retry_count after x seconds of successful running
        self.reset_retry_threshold = parse_duration_literal(reset_retry_threshold)
        self.last_error = None  # type: Optional[datetime]

    def handle_error(self) -> RetryDirective:
        # Handles two cases:
        # 1. Initial value -> no retries so far, the next is 1
        # 2. Reset retry count because threshold has reached, next try is 1
        if (self.last_error is None
                or (datetime.now() - self.last_error).total_seconds() > self.reset_retry_threshold):
            self.retry_count = 0

        self.last_error = datetime.now()
        directive = super().handle_error()

        return RetryDirective(
            abort=directive.abort,
            wait_for=self.retry_wait,
            retry_cnt=self.retry_count
        )


class PushExecutor(Loggable, Singleton):
    """
    Given a payload and the push this helper actually "executes" the push by passing the payload
    after the (optional) selector did some magic to it.

    Examples:

        >>> payload = dict(a="This is the payload", b="another ignored key by selector")
        >>> from pnp.plugins.push.simple import Nop
        >>> push_instance = Nop(name='doctest')
        >>> push = PushModel(instance=push_instance, selector='data.a', deps=[], unwrap=False)
        >>> PushExecutor().execute("doctest", payload, push)
        >>> push_instance.last_payload
        'This is the payload'

    """

    def _execute_internal(self, ident: str, payload: Payload, push: PushModel,
                          result_callback: Optional[PushResultCallback] = None) -> None:
        self.logger.debug("[%s] Selector: Applying '%s' to '%s'", ident, push.selector, payload)
        payload = PayloadSelector.instance.eval_selector(  # pylint: disable=no-member
            push.selector,
            copy.deepcopy(payload)
        )

        # Only make the push if the selector wasn't evaluated to suppress the push
        if payload is not PayloadSelector.instance.suppress:  # pylint: disable=no-member
            self.logger.debug("[%s] Emitting '%s' to push '%s'", ident, payload, push.instance)
            push_result = push.instance.push(payload=payload)

            # Trigger any dependent pushes
            for dependency in push.deps:
                if result_callback:
                    # Delegate work back to the engine
                    self.logger.debug(
                        "[%s] Dependency callback is given. Delegating work back to engine", ident
                    )
                    result_callback(push_result, dependency)
                else:
                    # No delegation callback defined: The executor will recursively call itself
                    self.logger.debug(
                        "[%s] No callback is given. Recursively process dependencies", ident
                    )
                    self.execute(ident, push_result, dependency)
        else:
            self.logger.debug(
                "[%s] Selector evaluated to suppress literal. Skipping the push", ident
            )

    def execute(self, ident: str, payload: Payload, push: PushModel,
                result_callback: Optional[PushResultCallback] = None) -> None:
        """
        Executes the given push by passing the specified payload.
        In concurrent environments there might be multiple executions in parallel.
        You may specify an `id` argument to identify related execution steps in the logs.
        Use the `result_callback` when the engine can take care of dependent pushes as well.
        The result and a dependent push will be passed via the callback. If the callback is not
        specified the PushExecute will execute them in a recursive manner.

        Args:
            ident (str): ID to identify related execution steps in the logs
                (makes sense in concurrent environments).
            payload (Any): The payload to pass to the push.
            push (PushModel): The push instance that has to process the payload.
            result_callback (callable): See explanation above.

        Returns:
            None.
        """
        Validator.is_instance(PushModel, push=push)

        if result_callback and not callable(result_callback):
            self.logger.warning(
                "Result callback is given, but is not a callable. Callback will be ignored."
            )
            result_callback = None

        if push.unwrap and is_iterable_but_no_str(payload):
            length = len(payload)
            self.logger.debug("[%s] Unwrapping payload to %s individual items", ident, str(length))
            for item in payload:
                self._execute_internal(ident, item, push, result_callback)
        else:
            self._execute_internal(ident, payload, push, result_callback)



"""Data model."""

from typing import Dict, List, Optional, Any, cast, Iterable, Union

import attr
from box import Box  # type: ignore

from .plugins import load_plugin
from .plugins.pull import PullBase
from .plugins.push import PushBase
from .plugins.udf import UserDefinedFunction
from .typing import AnyCallable, SelectorExpression


def _val_deps(instance, attrib, val):  # type: ignore  # pylint: disable=unused-argument
    if val is None:
        return None
    if not isinstance(val, list):
        raise TypeError("Dependencies is expected to be a list of pushes")
    for exp_push in val:
        if not isinstance(exp_push, PushModel):
            raise TypeError("A dependency is expected to be a child of pnp.plugins.push.PushBase")
    return val


def _val_pushes(instance, attrib, val):  # type: ignore  # pylint: disable=unused-argument
    if not isinstance(val, list):
        raise TypeError("Pushes is expected to be a list of pushes")
    for exp_push in val:
        if not isinstance(exp_push, PushModel):
            raise TypeError("A push is expected to be a child of pnp.models.Push")


def _val_callable(instance, attrib, val):  # type: ignore  # pylint: disable=unused-argument
    if not callable(val):
        raise TypeError("The value is expected to be a callable")


@attr.s
class PullModel:
    """Model representing a pull."""
    instance = attr.ib(validator=attr.validators.instance_of(PullBase))  # type: PullBase


@attr.s
class PushModel:
    """Model representing a push."""
    instance = attr.ib(validator=attr.validators.instance_of(PushBase))  # type: PushBase
    selector = attr.ib(validator=attr.validators.optional(
        attr.validators.instance_of((str, list, dict))  # type: ignore
    ), default=None)  # type: Optional[SelectorExpression]
    unwrap = attr.ib(converter=bool, default=False)  # type: bool
    deps = attr.ib(validator=_val_deps, default=list())  # type: List[PushModel]


@attr.s
class TaskModel:
    """Model representing a task (pull and dependant pushes)."""
    name = attr.ib(converter=str)  # type: str
    pull = attr.ib(validator=attr.validators.instance_of(PullModel))  # type: PullModel
    pushes = attr.ib(validator=_val_pushes)  # type: List[PushModel]

    @classmethod
    def mk_pull(cls, task: Box, **extra: Any) -> PullModel:
        """Make a pull out of a task configuration."""
        args = {'name': '{task.name}_pull'.format(task=task), **extra, **task.pull.args}
        return PullModel(instance=cast(PullBase, load_plugin(
            plugin_path=task.pull.plugin,
            plugin_type=PullBase,
            instantiate=True,
            **args
        )))

    @classmethod
    def mk_push(cls, task: Box, **extra: Any) -> List[PushModel]:
        """Make one or more pushes out of task configuration."""
        def _many(pushlist: List[Box], prefix: str) -> Iterable[PushModel]:
            for i, push in enumerate(pushlist):
                push_name = '{prefix}_{i}'.format(
                    i=i,
                    prefix=prefix
                )
                args = {'name': push_name, **extra, **push.args}
                unwrap = getattr(push, 'unwrap', False)
                yield PushModel(
                    instance=cast(PushBase, load_plugin(
                        plugin_path=push.plugin,
                        plugin_type=PushBase,
                        instantiate=True,
                        **args
                    )),
                    selector=push.selector,
                    unwrap=unwrap,
                    deps=list(_many(push.deps, push_name))
                )
        return list(_many(task.pushes, "{task.name}_push".format(**locals())))

    @classmethod
    def from_dict(cls, task: Box, base_path: Optional[str] = None) -> 'TaskModel':
        """Create a task from a task configuration."""
        if not isinstance(task, Box):
            task = Box(dict(base=task)).base

        extra = {'base_path': base_path}
        return cls(
            name=task.name,
            pull=cls.mk_pull(task, **extra),
            pushes=list(cls.mk_push(task, **extra))
        )


@attr.s
class UDFModel:
    """Model representing a user-defined function."""
    name = attr.ib(converter=str)  # type: str
    callable = attr.ib(validator=_val_callable)  # type: Union[AnyCallable, UserDefinedFunction]

    @classmethod
    def from_config(cls, dct: Union[Box, Dict[Any, Any]]) -> 'UDFModel':
        """Creates a UDFModel from a udf configuration snippet."""
        if not isinstance(dct, Box):
            dct = Box(dct)
        udf_type = cast(Union[str, type],
                        'callable' if not hasattr(dct, 'args') else UserDefinedFunction)
        instantiate = hasattr(dct, 'args')
        kwargs = dict() if not hasattr(dct, 'args') or dct.args is None else dct.args
        fun = load_plugin(
            plugin_path=dct.plugin,
            plugin_type=udf_type,
            instantiate=instantiate,
            **{'name': dct.name, **kwargs}
        )
        return cls(name=dct.name, callable=cast(UserDefinedFunction, fun))


TaskSet = Dict[str, TaskModel]


def tasks_to_str(tasks: TaskSet) -> str:
    """Transforms a set of tasks to a human readable representation."""
    res = []
    for _, task in tasks.items():
        res.append('{task.name} {{'.format(task=task))
        res.append('\tpull = {task.pull}'.format(task=task))
        res.append('\tpushes = ['.format())
        for push in task.pushes:
            res.append('\t\t{push}'.format(push=push))
        res.append('\t]')
        res.append('}')
    return "\n".join(res)



"""Pull 'n' Push

Usage:
  pnp [(-c | --check)] [(-v | --verbose)] [--log=<log_conf>] <configuration>
  pnp (-h | --help)
  pnp --version

Options:
  -c --check        Only check configuration and do not run it.
  -v --verbose      Switches log level to debug.
  --log=<log_conf>  Specify logging configuration to load.
  -h --help         Show this screen.
  --version         Show version.
"""

import logging
import logging.config
import os

from docopt import docopt
from ruamel import yaml
from schema import Schema, Use, And, Or

from ..app import Application
from ..utils import get_first_existing_file


def _setup_logging(*candidates, default_level=logging.INFO, env_key='PNP_LOG_CONF', verbose=False):
    """Setup logging configuration"""
    log_file_path = get_first_existing_file(*candidates)
    env_path = os.getenv(env_key, None)
    if env_path:
        log_file_path = env_path
    if log_file_path and os.path.exists(log_file_path):
        with open(log_file_path, 'rt') as fhandle:
            config = yaml.safe_load(fhandle.read())
        logging.config.dictConfig(config)
        logging.info("Logging loaded from: %s", log_file_path)
        if verbose:
            logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                            level=logging.DEBUG if verbose else default_level)
        logging.info("Logging loaded with basic configuration")


def _validate_args(args):
    validator = Schema({
        "--help": Use(bool),
        "--verbose": Use(bool),
        "--version": Use(str),
        "--check": Use(bool),
        "--log": Or(None, And(os.path.isfile, Use(os.path.abspath))),
        "<configuration>": And(os.path.isfile, Use(os.path.abspath))
    })
    return validator.validate(args)


def run(args):
    """Run pull 'n' push."""
    validated = _validate_args(args)
    pnp_cfg_path = validated['<configuration>']
    default_log_level = os.environ.get('LOG_LEVEL', 'DEBUG')
    _setup_logging(
        args['--log'], 'logging.yaml', os.path.join(os.path.dirname(pnp_cfg_path), 'logging.yaml'),
        default_level=default_log_level, verbose=validated['--verbose']
    )
    app = Application.from_file(pnp_cfg_path)
    if not validated['--check']:
        app.start()


def main():
    """Main entry point into pnp application."""
    arguments = docopt(__doc__, version='0.17.0')
    run(arguments)


if __name__ == '__main__':
    main()



"""E-mail related push plugins."""

import base64
import os

from . import PushBase, enveloped, parse_envelope
from .. import load_optional_module
from ...shared.mime import Mail
from ...utils import FileLock, make_list, is_iterable_but_no_str
from ...validator import Validator


class GMail(PushBase):
    """
    Sends an e-mail via the gmail api.

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/mail.GMail/index.md
    """
    __prefix__ = 'gmail'
    EXTRA = 'gmail'

    def __init__(self, token_file, recipient, subject=None, sender='pnp', **kwargs):
        super().__init__(**kwargs)
        self.token_file = str(token_file)
        if not os.path.isabs(self.token_file):
            self.token_file = os.path.join(self.base_path, self.token_file)
        Validator.is_file(token_file=self.token_file)

        self.recipient = self._parse_recipient(recipient)
        self.subject = self._parse_subject(subject)
        self.sender = self._parse_sender(sender)

    @staticmethod
    def _create_message(sender, recipient, subject, message_text=None, attachment=None):
        if is_iterable_but_no_str(recipient):
            recipient = ",".join(recipient)
        if attachment is None:
            mail = Mail.create_text(sender, recipient, subject, message_text)
        else:
            mail = Mail.create_with_attachment(sender, recipient, subject, attachment, message_text)
        return {'raw': base64.urlsafe_b64encode(mail.as_bytes()).decode()}

    @staticmethod
    def _parse_recipient(val):
        recipients = make_list(val)
        return [str(v) for v in recipients]

    @staticmethod
    def _parse_subject(val):
        return val and str(val)

    @staticmethod
    def _parse_sender(val):
        return str(val)

    def _make_service(self, token_file):
        discovery = load_optional_module('googleapiclient.discovery', self.EXTRA)

        if not os.path.exists(token_file):
            raise RuntimeError("No credentials retrieved. Make sure '{}' exists".format(token_file))

        with FileLock(token_file):
            # The file token_file stores the user's access and refresh tokens
            import pickle
            with open(token_file, 'rb') as tokenfp:
                creds = pickle.load(tokenfp)

            # Check if the given credentials are still valid...
            # If no: Might be outdated so refresh them
            if not creds.valid:
                if creds.expired and creds.refresh_token:
                    requests = load_optional_module('google.auth.transport.requests', self.EXTRA)
                    creds.refresh(requests.Request())
                    # Save the credentials for the next run
                    import pickle
                    with open(token_file, 'wb') as tokenfp:
                        pickle.dump(creds, tokenfp)
                else:
                    raise RuntimeError("Tokens are not valid and cannot be refreshed! Period.")

        return discovery.build('gmail', 'v1', credentials=creds, cache_discovery=False)

    def _send_message(self, service, message, user_id='me'):
        """Send an email message.

        Args:
            service: Authorized Gmail API service instance.
            message: Message to be sent.
            user_id: User's email address. The special value "me"
            can be used to indicate the authenticated user.

        Returns:
            Sent Message.
        """
        try:
            return service.users().messages().send(userId=user_id, body=message).execute()
        except:  # pylint: disable=bare-except
            import traceback
            self.logger.error("Error: %s", traceback.format_exc())

    @enveloped
    @parse_envelope('recipient')
    @parse_envelope('subject')
    @parse_envelope('sender')
    def push(self, recipient, subject, sender, envelope, payload):  # pylint: disable=arguments-differ
        file_path = envelope.get('attachment')

        if subject is None:
            raise ValueError("Subject was not defined either by the __init__ nor by the envelope")
        Validator.is_file(allow_none=True, file_path=file_path)

        self.logger.info(
            "Sending E-Mail '%s' to %s\n"
            "Content: '%s' (Attachment: '%s')",
            subject, recipient, payload, file_path
        )

        message = self._create_message(sender, recipient, subject, payload, file_path)
        service = self._make_service(self.token_file)
        self._send_message(service, message)

        return {'data': payload, **envelope} if envelope else payload



"""Contains file system related push plugins."""

import os
import time

from . import PushBase, enveloped, parse_envelope, drop_envelope
from ...validator import Validator


class FileDump(PushBase):
    """
    This push dumps the given `payload` to a file to the specified `directory`.
    If argument `file_name` is None, a name will be generated based on the current datetime
    (%Y%m%d-%H%M%S).
    If `file_name` is not passed (or None) you should pass `extension` to specify the extension of
    the generated file name.
    Argument `binary_mode` controls whether the dump is binary (mode=wb) or text (mode=w).


    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/fs.FileDump/index.md

    Examples:

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     # Automatic generated file name with extension '.dump'
        ...     dut = FileDump(name='doctest', directory=tmpdir, binary_mode=False)
        ...     created_file = dut.push("I am the content")  #  e.g. 20180616-123159.dump
        ...     with open(created_file, 'r') as fs:
        ...         assert fs.read() == "I am the content"

    """
    def __init__(self, directory='.', file_name=None, extension='.dump', binary_mode=True,
                 **kwargs):
        super().__init__(**kwargs)
        self.directory = directory
        Validator.is_directory(directory=self.directory)
        self.extension = self._parse_extension(extension)
        self.binary_mode = bool(binary_mode)
        self.file_name = self._parse_file_name(file_name)

    @staticmethod
    def _parse_file_name(value):
        return Validator.cast_or_none(str, value)

    @staticmethod
    def _parse_extension(value):
        value = str(value)
        if not value.startswith('.'):
            return'.' + value
        return value

    @enveloped
    @parse_envelope('file_name')
    @parse_envelope('extension')
    @drop_envelope
    def push(self, file_name, extension, payload):  # pylint: disable=arguments-differ
        if file_name is None:
            file_name = time.strftime("%Y%m%d-%H%M%S")
        file_path = os.path.join(self.directory, file_name + extension)
        with open(file_path, 'wb' if self.binary_mode else 'w') as fhandle:
            fhandle.write(str(payload) if not self.binary_mode else payload)

        return file_path



"""Configuration handling in json/yaml files."""

import os
from functools import partial

from ruamel import yaml  # type: ignore
from schema import Schema, Use, Optional, Or, And  # type: ignore
from box import Box  # type: ignore

from .engines import Engine, RetryHandler
from .plugins import load_plugin
from .utils import make_list


class OrOverride(Or):
    """
    Validates the given data as the original Or validation directive would do,
    but overrides the resulting data.

    Example:
        >>> # Whether inbound or outbound are valid keys, but will be overridden by override
        >>> schema = Schema({OrOverride('override', 'inbound', 'outbound'): Use(str)})
        >>> schema.validate({'inbound': 'value'})  # inbound will be mapped to override
        {'override': 'value'}
        >>> schema.validate({'outbound': 'value'})  # outbound will be mapped to override
        {'override': 'value'}
        >>> OrOverride('override')  # Only override without valid directives is bad...
        Traceback (most recent call last):
        ...
        ValueError: OrMap expects the mapping value at first and then multiple validation directives

    """
    def __init__(self, *args, **kwargs):
        if len(args) < 2:
            raise ValueError("OrMap expects the mapping value at first and then "
                             "multiple validation directives")
        self.override = args[0]
        super().__init__(*args[1:], **kwargs)

    def validate(self, data):
        super().validate(data)
        return self.override


PUSH = Schema({
    "plugin": Use(str),
    Optional("selector", default=None): Or(object, None),
    Optional("unwrap", default=False): bool,
    Optional("args", default={}): {
        Optional(str): object
    },
    # deps is a list of push_schemas as well, but cannot declare recursive schemas
    Optional("deps", default=list()): And(object, Use(make_list))
})

PULL = Schema({
    "plugin": Use(str),
    Optional("args", default={}): {
        Optional(str): object
    },
})

PUSH_LIST = Schema([PUSH])

TASK = Schema({
    "name": Use(str),
    OrOverride("pull", "inbound", "pull"): PULL,
    OrOverride("pushes", "outbound", "push"): And(Or(PUSH_LIST, PUSH), Use(make_list))
})

# Allow configuration of a single task or a list of tasks
TASK_LIST = Schema(
    Or(
        And(TASK, Use(make_list)),
        [TASK]
    )
)

ENGINE = Engine

UDFS = Schema([{
    "name": Use(str),
    "plugin": Use(str),
    Optional("args"): Or({
        str: object
    }, None)
}])

TASK_SETTINGS = Schema({
    Optional(Or("anchors", "anchor", "ref", "refs", "alias", "aliases")): object,
    Optional("engine", default=None): ENGINE,
    Optional("udfs", default=None): UDFS,
    'tasks': TASK_LIST
})

ROOT = Schema(And(
    Or(
        # Tasks with additional settings and aliases (yaml only)
        TASK_SETTINGS,
        # Tasks without settings or aliases - just a list of tasks (backwards compatibility)
        TASK_LIST,
    ),
    # In case of list only we create a tasks node - so we know it's always there
    Use(lambda x: x if 'tasks' in x else dict(tasks=x))
))


def validate_push_deps(push):
    """Validates a push against the schema. This is done recursively."""
    for dependency in push['deps']:
        validated_push = PUSH.validate(dependency)
        validated_push['deps'] = list(validate_push_deps(validated_push))
        yield validated_push


def custom_constructor(loader, node, clstype):
    """YAML custom constructor. Necessary to create an engine and retry handler."""
    args = loader.construct_mapping(node, deep=True)
    if 'type' not in args:
        raise ValueError("You have to specify a 'type' when instantiating a engine with !engine")
    clazz_name = args.pop('type')
    return load_plugin(clazz_name, clstype, **args)


def make_mentor(config_path):
    """Creates an instance of the dictmentor with configured plugins."""
    from dictmentor import DictMentor, ext  # type: ignore
    return DictMentor(
        ext.Environment(fail_on_unset=True),
        ext.ExternalResource(base_path=os.path.dirname(config_path)),
        ext.ExternalYamlResource(base_path=os.path.dirname(config_path))
    )


def load_config(config_path):
    """Load the specified config"""
    yaml.SafeLoader.add_constructor(u"!engine", partial(custom_constructor, clstype=Engine))
    yaml.SafeLoader.add_constructor(u"!retry", partial(custom_constructor, clstype=RetryHandler))

    with open(config_path, 'r') as fp:
        cfg = yaml.safe_load(fp)
        # pnp configuration is probably of list of tasks. dictmentor needs a dict ...
        # ... let's fake it ;-)
        cfg = dict(cfg=cfg)

    mentor = make_mentor(config_path)
    # Remove the faked dictionary as root level

    validated = ROOT.validate(mentor.augment(cfg)['cfg'])
    for pull in validated['tasks']:
        for push in pull['pushes']:
            push['deps'] = list(validate_push_deps(push))
    b = Box(validated)

    return (
        b.udfs if 'udfs' in b else None,
        b.engine if 'engine' in b else None,
        b.tasks
    )



"""Engine implementation using a sequential workflow. Supports only one task."""

import multiprocessing
from typing import Optional

from . import (Engine, RetryHandler, PushExecutor, NotSupportedError, SimpleRetryHandler,
               NoRetryHandler)
from ..models import TaskSet, TaskModel
from ..typing import Payload
from ..utils import sleep_until_interrupt
from ..validator import Validator


class SequentialEngine(Engine):
    """Sequential engine that only allows to run a single pull. It is more like a script execution
    and nothing is concurrent."""
    def __init__(self, retry_handler: Optional[RetryHandler] = None):
        super().__init__()
        Validator.is_instance(RetryHandler, allow_none=True, retry_handler=retry_handler)
        self.retry_handler = retry_handler
        if not self.retry_handler:
            self.retry_handler = SimpleRetryHandler()
        self.stopped = multiprocessing.Event()
        self._task = None  # type: Optional[TaskModel]

    def _run(self, tasks: TaskSet) -> None:
        if not tasks:
            return
        if len(tasks) > 1:
            raise NotSupportedError("The Sequential engine can only handle one task")

        task = tasks[list(tasks.keys())[0]]
        self._task = task

        def on_payload(plugin: object, payload: Payload) -> None:  # pylint: disable=unused-argument
            for push in task.pushes:
                try:
                    PushExecutor().execute('sequential', payload, push)
                except:  # pylint: disable=bare-except
                    import traceback
                    self.logger.error("\n%s", traceback.format_exc())

        task.pull.instance.on_payload = on_payload  # type: ignore
        task.pull.instance._stopped = self.stopped  # pylint: disable=protected-access

        while not self.stopped.is_set():
            try:
                task.pull.instance.pull()
                if isinstance(self.retry_handler, NoRetryHandler):
                    self.stopped.set()

                if not self.stopped.is_set():
                    # Bad thing... Pulling exited unexpectedly
                    self.logger.error(
                        "Pulling of '%s' exited unexpectedly",
                        task.pull.instance,
                    )
            except KeyboardInterrupt:
                self.stopped.set()
            except:   # pylint: disable=bare-except
                # Bad thing... Pulling exited with exception
                import traceback
                self.logger.error(
                    "Pulling of '%s' raised an error\n%s",
                    task.pull.instance,
                    traceback.format_exc()
                )
            finally:
                if not self.stopped.is_set():
                    directive = self.retry_handler.handle_error()  # type: ignore
                    if directive.abort:
                        self.logger.error(
                            "Pulling of '%s' exited due to retry limitation",
                            task.pull.instance,
                        )
                        self.stopped.set()
                    else:
                        sleep_until_interrupt(directive.wait_for, self.stopped.is_set)

    def _stop(self) -> None:
        self.stopped.set()



"""Pull 'n' Sound recorder

Usage:
  pnp_record_sound <output_file> <seconds> [--index=<idx>]
  pnp_record_sound --list
  pnp_record_sound (-h | --help)

Options:
  -h --help         Show this screen.
  --list            Shows all available mic devices and their associated index

Args:
  output_file       File to write (wav)
  seconds           Number of seconds to record
  --index=<idx>     Optionally specify the index of the mic to record the sound

"""

CHANNELS = 1
RATE = 44100
CHUNK = 512


def run(out_wav, secs, idx):
    """
    Run the sound recorder.

    Args:
        out_wav (str): File path where to store the recorded sound.
        secs (int): Number of seconds to record.
        idx (Optional[int]): The device index to use or None (first qualified device).
    """
    try:
        import pyaudio
        import wave
    except ImportError:
        print("You have to install extra 'sound' in order to use this shell script")
        return 99

    fmt = pyaudio.paInt16
    pa = pyaudio.PyAudio()
    if idx is not None and idx > pa.get_device_count() - 1:
        print("Device index is out of bounds")
        return 1
    if idx is not None:
        device = pa.get_device_info_by_index(idx)
        if device.get('maxInputChannels', 0) == 0:
            print("Device @ index {idx} is not an input device (microphone)".format(**locals()))
            return 2
        if device.get('maxInputChannels', 0) < CHANNELS:
            print("Device @ index {idx} does not support {channels} channels".format(
                idx=idx, channels=CHANNELS
            ))
            return 3

    buffer = []
    input("Press any key to start the recording")
    stream = pa.open(
        format=fmt,
        channels=CHANNELS,
        rate=RATE,
        input=True,
        input_device_index=idx,
        frames_per_buffer=CHUNK
    )

    try:
        for _ in range(0, int(RATE / CHUNK * secs)):
            data = stream.read(CHUNK)
            buffer.append(data)
    finally:
        stream.stop_stream()
        stream.close()
        pa.terminate()

    out_file = wave.open(out_wav, 'wb')
    out_file.setnchannels(CHANNELS)
    out_file.setsampwidth(pa.get_sample_size(fmt))
    out_file.setframerate(RATE)
    out_file.writeframes(b''.join(buffer))
    out_file.close()

    print("Wave file written to {out_wav} @ {rate} Hz".format(out_wav=out_wav, rate=RATE))
    return 0


def list_devices():
    """List all available microphone devices."""
    try:
        import pyaudio
    except ImportError:
        print("You have to install extra 'sound' in order to use this shell script")
        return 99

    pa = pyaudio.PyAudio()
    for i in range(pa.get_device_count()):
        dev = pa.get_device_info_by_index(i)
        input_chn = dev.get('maxInputChannels', 0)
        if input_chn > 0:
            name = dev.get('name')
            rate = dev.get('defaultSampleRate')
            print("Index {i}: {name} (Max Channels {input_chn}, Default @ {rate} Hz)".format(
                i=i, name=name, input_chn=input_chn, rate=int(rate)

            ))
    return 0


def main():
    """Main entry point to runner pnp_record_sound."""
    from docopt import docopt
    arguments = docopt(__doc__)
    if arguments.get('--list', False):
        return list_devices()

    out_wav, secs, idx = (
        arguments['<output_file>'],
        int(arguments['<seconds>']),
        arguments.get('--index', None)
    )
    return run(out_wav, secs, idx and int(idx))


if __name__ == '__main__':
    import sys
    sys.exit(main())



"""Notification related push plugins."""

from . import PushBase, enveloped
from .. import load_optional_module
from ...utils import auto_str_ignore


@auto_str_ignore(['api_key'])
class Pushbullet(PushBase):
    """
    Sends a message to the Pushbullet service. The type of the message will guessed:

    - `push_link` for a single http link
    - `push_file` if the link is directed to a file (mimetype will be guessed)
    - `push_note` for everything else (converted to str)

    See Also:
        https://github.com/HazardDede/pnp/blob/master/docs/plugins/push/notify.Pushbullet/index.md
    """
    __prefix__ = 'pushbullet'

    EXTRA = 'pushbullet'

    def __init__(self, api_key, title='pnp', **kwargs):
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.title = str(title)

    def _load_deps(self):
        pbullet = load_optional_module('pushbullet', self.EXTRA)
        try:
            import urlparse
        except:  # For Python 3, pylint: disable=bare-except
            import urllib.parse as urlparse

        return pbullet, urlparse

    def _guess_mimetype(self, url):
        import mimetypes
        _, urlparse = self._load_deps()
        res = urlparse.urlparse(url)
        if not res.path:
            return None
        mtype, _ = mimetypes.guess_type(res.path)
        return mtype

    @staticmethod
    def _safe_basename(name):
        import os
        try:
            return os.path.basename(name)
        except:  # pylint: disable=bare-except
            return name

    @enveloped
    def push(self, envelope, payload):  # pylint: disable=arguments-differ
        pbullet, urlparse = self._load_deps()
        client = pbullet.Pushbullet(self.api_key)
        urlprofile = urlparse.urlparse(payload)
        self.logger.info("Sending message '%s' to Pushbullet", payload)
        if urlprofile.scheme and urlprofile.netloc:
            # Is URL
            self.logger.debug("Payload '%s' is an url", payload)
            mtype = self._guess_mimetype(payload)
            if not mtype:
                # Some link
                self.logger.debug(
                    "Payload '%s' has no mimetype associated", payload
                )
                client.push_link(self.title, payload)
            else:
                # Some file
                self.logger.debug(
                    "Guessed type '%s' for payload '%s'", mtype, payload
                )
                client.push_file(
                    file_name=self._safe_basename(urlprofile.path),
                    file_type=mtype,
                    file_url=payload,
                    title=self.title
                )
        else:
            # Is NOT url
            self.logger.debug("Payload '%s' is a simple note", payload)
            client.push_note(self.title, payload)

        return {'data': payload, **envelope} if envelope else payload
